classdef PulseStreamer < handle
    % PulseStreamer is a wrapper class for the JSON-RPC interface of the Swabian Instruments' Pulse Streamer
    %
    % https://www.swabianinstruments.com/static/documentation/PulseStreamer/sections/api-doc.html#pulsestreamer
    %
    
    properties (Constant)
        REPEAT_INFINITELY = -1;  % Constant for infinite repetitions of the sequence
        SUPPORT_EMAIL = 'support@swabianinstruments.com'; % Support contact email, for display only.
    end
    
    properties (SetAccess = private, GetAccess = public)
        ip                       % IP address of the Pulse Streamer
        analog_channels = 0:1;   % List of valid analog channels for the connected Pulse Streamer
        digital_channels = 0:7;  % List of valid analog channels for the connected Pulse Streamer
        hasFinishedCallbackFcn   % callback function called when the stream of the Pulse Streamer has finished
    end
    
    properties (Access = {?PulseStreamer.PulseStreamer})
        max_step_duration = 4294967295; % Maximum duration of a sequence step (uint32). Longer durations will be split.
        max_step_count = 2e6;    % maximum number of steps in a sequence
        analog_scale = 32767;    % Scaling factor that converts physical value to ADC value
        
        cb_timer_name = '';      % callback polling timer name
        sequence_duration        % length of the sequence in ns including multiple runs
        n_runs                   % store internally the number of runs
        final_output             % store internally final output value
        fw_ver;
        has_streamBinary = true; % has capability to stream binary
        can_streamBinary = true; % true if binary streaming is working properly
    end
    
    methods
        % constructor
        function obj = PulseStreamer(ip)
            % Class constructor
            % ip - is an IP address of the pulse streamer (e.g.
            % 'pulsestreamer' or '169.254.8.2')
            % You can specify empty string as IP address for offline sequence design.
            
            obj.ip = ip;
            
            % Check if we are able to communicate to Pulse Streamer
            try
                obj.httpRequest('{"jsonrpc":"2.0","id":"1","method":"getFirmwareVersion"}');
            catch e
                showChars = min(50, length(e.message));
                error(['Could not connect to Pulse Streamer at "%s". \n\n' ...
                     'Error message:\n %s (...)'], ip, e.message(1:showChars));
            end
            
            % Check firmware version compatibility
            obj.checkFirmwareCompatibility();
            
            % Generate callback timer name string based on device serial number
            % The name string is derived from the Pulse Streamer serian number.
            obj.cb_timer_name = obj.getSerial();
            
            % Check if associated callback timer exists and remove it.
            obj.stopCallbackTimer(); 
        end
        
        function sequence = createSequence(obj)
            % Creates Sequence object specific to the hardware model
            
            % switch (obj.model)
            % case 'PulseStreamer 8/2'
            sequence = PulseStreamer.hw.PS_D8A2.Sequence();
            % 
            % otherwise
            %   sequence = PulseStreamer.Sequence();
            % end
        end
        
        %%%%%%%%%%%%%%%%%%%% wrapped JSON-RPC methods %%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        function reset(obj)
           % resets the Pulse Streamer device
           obj.stopCallbackTimer();
           obj.callJSONRPC('reset');
        end
        
        function constant(obj, state)
            % set a constant output at the Pulse Streamer outputs
            obj.stopCallbackTimer();
            
            enc_state = obj.outputStateEncode(state);
            obj.callJSONRPC('constant', enc_state);
        end
    
        function stream(obj, sequence, n_runs, final)
            % STREAM sends the sequence to the Pulse Streamer
            %  sequence:  Sequence object or cell array.
            %  n_runs:    Number of times to repeat the sequence.
            %             Infinite repetitions if n_runs<0.
            %  final:     OutputState object which defines the output
            %             after the sequence has ended
            
            if ~exist('final', 'var')
                final = PulseStreamer.OutputState.ZERO;
            end
            
            if ~exist('n_runs', 'var')
                n_runs = PulseStreamer.PulseStreamer.REPEAT_INFINITELY;
            end
            
            if isempty(sequence)
                % Do nothing when sequence is empty
                warning('Sequence is empty. Setting final state\n');
                obj.constant(final); % Set final state immediately
                return;
            end

            obj.stopCallbackTimer();
            
            rle = obj.sequenceConvertToRLEData(sequence);            
            obj.streamRLEData(rle, n_runs, final);
        end
        
        function TF = rearm(obj)
            % Rearm Pulse Streamer
            ret = obj.callJSONRPC('rearm');
            TF = jsonToBool(ret);
        end
        
        function startNow(obj)
            % starts the sequence if the sequence was uploaded with the
            % PSStart.Software option
            obj.callJSONRPC('startNow');
        end
        
        function forceFinal(obj)
            % Interrupts the sequence and sets the final state. This
            % method does not modify the output state if the sequence has 
            % already finished and the Pulse Streamer was in the final 
            % state. This method also releases the hardware resources of  
            % the Pulse Streamer and, therefore, allows for faster upload
            % sequence during next call of "stream" method.
            
            obj.callJSONRPC('forceFinal');
        end  
        
        function setTrigger(obj, start, rearm)
            %SETTRIGGER: Defines how the uploaded sequence is triggered.
            
            if ~isa(start,'PulseStreamer.TriggerStart')
                error('Invalid type of "start" parameter. Use TriggerStart enumeration.');
            end
            if exist('rearm', 'var')
                if ~isa(rearm, 'PulseStreamer.TriggerRearm')
                    error('Invalid type of "rearm" parameter. Use TriggerRearm enumeration.');
                end
            else
                rearm = PulseStreamer.TriggerRearm.AUTO;
            end
            obj.callJSONRPC('setTrigger', start, rearm);
        end
        
        function source = getTriggerStart(obj)
            % Return current trigger start mode
            ret = obj.callJSONRPC('getTriggerStart');
            response = jsonToStruct(ret);
            source = PulseStreamer.TriggerStart(response.result);
        end
        
        function source = getTriggerRearm(obj)
            % Return current trigger rearm mode
            ret = obj.callJSONRPC('getTriggerRearm');
            response = jsonToStruct(ret);
            source = PulseStreamer.TriggerRearm(response.result);
        end
        
        function TF = isStreaming(obj)
            % check whether the Pulse Streamer is currently outputting a sequence
            ret = obj.callJSONRPC('isStreaming');
            TF = jsonToBool(ret);
        end
        
        function TF = hasSequence(obj)
            % check whether a sequence was uploaded to 
            ret = obj.callJSONRPC('hasSequence');
            disp(ret)
            TF = jsonToBool(ret);
        end
        
        function TF = hasFinished(obj)
            % check whether all sequences are finished
            ret = obj.callJSONRPC('hasFinished');
            TF = jsonToBool(ret);
        end
        
        function selectClock(obj, source)
            % Select Clock source type
            if ~isa(source, 'PulseStreamer.ClockSource')
                error('Invalid type of clockSource. Use ClockSource enumeration.');
            end
            obj.callJSONRPC('selectClock', source);
        end
        
        function source = getClock(obj)
            % Return current clock source
            ret = obj.callJSONRPC('getClock');
            response = jsonToStruct(ret);
            source = PulseStreamer.ClockSource(response.result);
        end
        
        function serial_str = getSerial(obj)
            %getSerial: Returns serial number as string 
            %
            %   The serial numer is the same as MAC address of the ethernet
            %   interface.  
            
            if strncmp(obj.fw_ver, '1.0', 3)
                % COMPATIBILITY: Firmware v1.0 expects getSerial with a parameter.
                ret = obj.callJSONRPC('getSerial', 1);
            else
                ret = obj.callJSONRPC('getSerial');
            end
            
            response = jsonToStruct(ret);
            serial_str = response.result;
        end
        
        function fpga_id_str = getFPGAID(obj)
            %getSerial: Returns FPGA ID number as string 
            %
            
            try
                ret = obj.callJSONRPC('getFPGAID');    
            catch
                if strncmp(obj.fw_ver, '1.0', 3)
                    % COMPATIBILITY: Firmware v1.0 expects getSerial with a parameter.
                    ret = obj.callJSONRPC('getSerial', 0);
                else
                    rethrow(e)
                end
            end
            
            response = jsonToStruct(ret);
            fpga_id_str = response.result;
        end
        
        function fw_ver = getFirmwareVersion(obj)
            %getFirmwareVersion: Returns firmware version string of the
            %connected Pulse Streamer.
            
            ret = obj.callJSONRPC('getFirmwareVersion');
            response = jsonToStruct(ret);
            fw_ver = response.result;
        end
        
        function hostname = getHostname(obj)
            %getHostname: Returns Pulse Streamer's hostname
            %
            ret = obj.callJSONRPC('getHostname');
            response = jsonToStruct(ret);
            hostname = response.result;
        end
        
        function setHostname(obj, hostname)
            %setHostname: Sets Pulse Streamer's hostname
            %
            obj.callJSONRPC('setHostname', hostname);
        end
        
        %%%%%%%% callback function and event handling %%%%%%%%%%%%%%%%%%%%%
        function setCallbackFinished(obj, func)
            % sets the callback function to detect when the Pulse Streamer
            % is finished with all sequences
            % this must be set before the the sequence is started.
            % e.g. fun: @myCallbackFunction
            %
            % the signature of the callback function must be
            % callbackFunction(pulseStreamer)
            
            % check whether the parameter is valid
            if ~isa(func,'function_handle')
                error('Callback function must be a function handle');
            end
            
            if exist(func2str(func)) == 0
                error(['Callback function not found. ' func2str(func)]);
            end
            obj.hasFinishedCallbackFcn = func;
        end
        
    end
    
    methods
       % Overriden methods. 
       function delete(obj)
           % class destructor
           % Stop and delete timer on object deletion.
           % This method exists only in handle classes, 
           % i.e. ones subclassed from "handle" superclass
           
           % delete(h) deletes a handle object, but does not clear the
           % handle variable from the workspace. The handle variable is
           % not valid once the handle object has been deleted.
           %
           % https://uk.mathworks.com/help/matlab/matlab_oop/handle-class-destructors.html
           %
           % To be a valid class destructor, the delete method:
           %   *  Must define one, scalar input argument, which is an object of the class.
           %   *  Must not define output arguments
           %   *  Cannot be Sealed, Static, or Abstract
           % In addition, the delete method should not:
           %   *  Throw errors, even if the object is invalid.
           %   *  Create new handles to the object being destroyed
           %   *  Call methods or access properties of subclasses
           
           if isempty(obj.cb_timer_name)
               %Find timer by the name 
               tobj = timerfind('Name', obj.cb_timer_name);
               if isa(tobj, 'timer')
                   if tobj.isvalid
                       stop(tobj);
                   end
                   delete(tobj);
               end
               obj.cb_timer_name = '';
           end
           
           % Defining a delete method in a handle subclass
           % does not override the handle class delete method.
           % No need to call delete@handle(obj) ourselves!!!
       end
    end

    % Private methods. Accessible only by direct subclasses.
    methods (Access = {?PulseStreamer.PulseStreamer})    
        
        function rle = sequenceConvertToRLEData(obj, seq)
            % Function takes sequence input and normalizes it to a common
            % RLEData format.
            
            % Fully qualified class name
            class_fqn = split(class(seq), '.');
            
            switch class_fqn{end} % class name without package names
                case {'Sequence', 'PulseStreamer.Sequence'}
                    % input is a Sequence object
                    rle = seq.getData();
                    
                case {'RLEData', 'PulseStreamer.internal.RLEData'}
                    % input is an RLEData object
                    rle = seq;
                    
                case {'OutputState', 'PulseStreamer.OutputState'}
                    % input is an OutputState object
                    % Convert to RLEData with a single step
                    rle = PulseStreamer.internal.RLEData();
                    rle.durations = 0;
                    if ~isempty(seq.chans)
                        rle.dchans = seq.chans;
                        rle.dlvls(1, :) = ones(size(seq.chans));
                    end
                    rle.achans = [0, 1];
                    rle.alvls(1, 1:2) = [seq.ao0, seq.ao1];
                    
                case 'cell'
                    % input is a cell array
                    % Example:
                    %   {duration_1, channels_1, A0_1, A1_1;
                    %     ...
                    %    duration_N, channels_N, A0_N, A1_N}
                    %
                    
                    [n_steps, n_col] = size(seq);
                    
                    if n_col<2
                        % Analog output values are optional
                        error(['Cell array must be of the format\n' ...
                              ' {duration_1, channels_1, A0_1, A1_1;\n' ...
                              ' ...\n' ...
                              ' duration_N, channels_N, A0_N, A1_N}']);
                    end
                    
                    rle = PulseStreamer.internal.RLEData();
                    rle.durations = cell2mat(seq(:, 1));
                    rle.dchans = obj.digital_channels;
                    rle.dlvls = zeros(n_steps, numel(obj.digital_channels));
                    
                    for ii = 1:n_steps
                        msk = ismember(obj.digital_channels, seq{ii, 2});
                        rle.dlvls(ii, :) = msk;
                    end
                    
                    if n_col > 2
                        rle.achans = 0:(n_col - 3);
                        rle.alvls = cell2mat(seq(:, 3:n_col));
                    end
                    
                otherwise
                    error('Unable to parse sequence data type: %s', class(seq));
            end
            
            % Map channels in RLE to the PulseStreamer channels
            rle = obj.sequenceMapChannels(rle);
        end
        
        
        function rle = sequenceMapChannels(obj, rle)
            % This method maps the RLE channels to the physical output
            % channels of the PulseStreamer.
            % Raises error when data contains channel numbers that are not
            % present on the device. 
            
            % Map digital channels to PulseStreamer channels
            [msk, ps_map] = ismember(rle.dchans, obj.digital_channels);
            if ~all(msk) 
                no_chans = num2str(obj.digital_channels(~msk));
                error('Digital channel(s) [%s] is(are) not present on the device!', no_chans);
            end
            dlvls = zeros(numel(rle.durations), numel(obj.digital_channels));
            dlvls(:, ps_map(msk)) = rle.dlvls;
            
            % Map analog channels to PulseStreamer channels
            [msk, ps_map] = ismember(rle.achans, obj.analog_channels);
            if ~all(msk)
                no_chans = num2str(obj.analog_channels(~msk));
                error('Analog channel(s) [%s] is(are) not present on the device!', no_chans);
            end
            alvls = zeros(numel(rle.durations), numel(obj.analog_channels));
            alvls(:, ps_map(msk)) = rle.alvls;
            
            % Put back to the same RLEData object
            rle.dchans = obj.digital_channels;
            rle.achans = obj.analog_channels;
            rle.alvls = alvls;
            rle.dlvls = dlvls;
        end      
  
                
        function [durations, digital, analog] = sequenceConvertToHWSpecific(obj, rle)
            % Convert Sequence RLE data to PulseStreamer specific RLE data
            % 
            % No type casting here otherwise 
            
            idxs = 1:numel(obj.digital_channels);
            powof2 = pow2(idxs - 1); % powers of 2
            
            durations = rle.durations;
            digital = rle.dlvls * powof2(:); % convert array of bits to decimal number;
            analog = round(rle.alvls .* obj.analog_scale);
        end


        function [durations, digital, analog] = sequenceSimplifySteps(obj, durations, digital, analog)
            % Simplifies sequence by combining steps with equal states and
            % removing steps with zero durations except the last.
        
            % Remove zero durations from data
            nz_msk = durations > 0;
            nz_msk(end) = true; % Keep last step even if duration=0
            durations = durations(nz_msk);  % Nonzero durations
            digital = digital(nz_msk, :); % Nonzero duration digital levels
            analog = analog(nz_msk, :); % Nonzero duration a0 levels
            
            % Find which values are repeated. 
            d_rep_msk = all(diff(digital, [], 1)==0, 2); % Repeated digital values
            a_rep_msk = all(diff(analog, [], 1)==0, 2); % Repeated analog values
            rep_msk = d_rep_msk & a_rep_msk; % Combined repetitions mask
            rep_msk = [false; rep_msk]; % Add the first element which is never a repetition.
            
            % Add up durations of repeated values.
            % For all repeated values, start from the end and add duration to previous
            % element. The first data element was already excluded before.
            rep_idxs = flipud(find(rep_msk(:))); % Start from the end!!!
            for ii = 1:numel(rep_idxs)
                idx = rep_idxs(ii);
                durations(idx-1) = durations(idx-1) + durations(idx);
            end
            
            % Write simplified data back to RLE object
            durations = durations(~rep_msk);
            digital = digital(~rep_msk, :);
            analog = analog(~rep_msk, :);
        end
        
        
        function [durations, digital, analog] = sequenceSplitLongSteps(obj, durations, digital, analog)
            % split long step durations that exceed maximum duration

            if any(durations > obj.max_step_duration)
                % Do splitting only if at least one step is too long
                
                dur_max = obj.max_step_duration;
                
                % Number of splits for each element
                % +1 to give space for remainders
                n_split = fix(durations ./ dur_max) + 1;
                
                % Indexes where to place remainders
                rem_idxs = cumsum(n_split); 
                
                % Data duplication indexes
                data_idxs = repelem(1:numel(n_split), n_split); 
                
                % Fill with "dur_max"
                dur_new = ones(1, sum(n_split)) .* dur_max; 
                
                % Replace remainders
                dur_new(rem_idxs) = durations - (n_split - 1) .* dur_max; 
                
                % Compose new states from original arrays
                durations = dur_new;
                digital = digital(data_idxs, :);
                analog = analog(data_idxs, :);
            end
        end
        
        
        function enc_state = outputStateEncode(obj, state)
            % Encodes OutputState object
            
            if isa(state, 'cell')
                % Create OutputState object from cell array
                state = PulseStreamer.OutputState(state{:});
            end
            
            % Handle as a single step sequence RLE data
            rle = obj.sequenceConvertToRLEData(state);
            
            % Rescale to hardware values
            [~, digi, anlg] = obj.sequenceConvertToHWSpecific(rle);
            
            % required parameter format
            enc_state = [0, digi(end), anlg(end, 1), anlg(end, 2)];
        end

        
        function streamRLEData(obj, rle, n_runs, final)
            % This is low-level function that sends RLE sequence data
            
            % Convert RLE data to hardware specific format
            [dur, digi, anlg] = obj.sequenceConvertToHWSpecific(rle);
                                    
            % Simplify steps by combining equal values and removing zero
            % durations
            [dur, digi, anlg] = obj.sequenceSimplifySteps(dur, digi, anlg);
            
            % Split long steps
            [dur, digi, anlg] = obj.sequenceSplitLongSteps(dur, digi, anlg);
            
            % Raise error when number of steps > max_step_count
            % Number of durations is the same as number of steps
            if numel(dur) > obj.max_step_count
                error(['Maximum number of sequence steps exceeded!' ...
                    'N_steps: %0.0f > max: %0.0f'], ...
                    numel(dur), obj.max_step_count);
            end
            
            % Send sequence data to the Pulse Streamer
            if obj.can_streamBinary
                obj.streamOverBinaryTCP(dur, digi, anlg, n_runs, final);
            else
                obj.streamOverJSON(dur, digi, anlg, n_runs, final);
            end 
        end
        
        
        function streamOverBinaryTCP(obj, dur, digi, anlg, n_runs, final)

            final_enc = obj.outputStateEncode(final);
            
            % Sequence data bytes
            seq_bytes = encodeBinary(dur, digi, anlg(:, 1), anlg(:, 2));
            
            % Create command object
            command = PulseStreamer.internal.cmdStreamHighLevel(...
                seq_bytes, n_runs, numel(dur), ...
                final_enc(3:4), final_enc(2)...
            );
            
            obj.callBinaryRPC(command);
            
            if (n_runs > 0) && isa(obj.hasFinishedCallbackFcn, 'function_handle')
                % the callback timer is only started if a callback function
                % is set by the user and as long the sequence has no infinite runs
                obj.startCallbackTimer();
            end
        end
        
        
        function streamOverJSON(obj, dur, digi, anlg, n_runs, final)
            % Encode sequence data into base64 string
            try
                error(javachk('jvm')); % check if Java is available
                
                % Use Java function to encode base64.
                % this is the fastest algorithm
                base64_seq = base64EncodeJava(dur, digi, anlg(:, 1), anlg(:, 2));
            catch ME
                % if java based base64 encoding fails then fallback to pure
                % MATLAB implementation
                if strcmp(ME.identifier, 'MATLAB:javachk:thisFeatureNotAvailable')
                    warning(['JVM is not available. ' ...
                        'Using pure MATLAB base64 encoder. ' ...
                        'Performance may be reduced.'])
                    base64_seq = base64EncodeMatlab(dur, digi,  anlg(:, 1), anlg(:, 2));
                else
                    rethrow(ME)
                end
            end
            
            obj.streamBase64EncodedSequence(base64_seq, n_runs, final);
        end
        
        
        function streamBase64EncodedSequence(obj, encoded_seq, n_runs, final)
            % This is low-level function that sends encoded sequence data
            
            obj.n_runs = n_runs;
            obj.final_output = final;
            enc_final = obj.outputStateEncode(final);
            
            obj.callJSONRPC('stream', encoded_seq, n_runs, enc_final);
            
            if (n_runs > 0) && isa(obj.hasFinishedCallbackFcn, 'function_handle')
                % the callback timer is only started if a callback function
                % is set by the user and as long the sequence has no infinite runs
                obj.startCallbackTimer();
            end
        end
        
        
        function resp = callBinaryRPC(obj, command)
            % Make binary encoded RPC call over TCP socket
            % command: Command object
            
            % Create reuqest object
            req = PulseStreamer.internal.Request(command);
            
            % Send request
            resp = obj.tcpRequest(req);
            
            if resp.status ~= PulseStreamer.internal.StatusEnum.OK
                error('ERROR code: %d, DATA:\n%s', ...
                    uint32(resp.status), evalc('disp(resp)'));
            end
        end
        
        
        function json_response = callJSONRPC(obj, method, varargin)
            % Make JSON-RPC call and wait for response
            % method:   name of the RPC call method
            % varagin:  arbitary number of arguments (string or numeric)

            persistent req_count;
            if isempty(req_count); req_count = uint32(0); end
            req_count = uint32(mod(req_count+1, intmax('uint32')));
            
            
            if nargin == 2
                json_request = sprintf('{"jsonrpc":"2.0","id":%d,"method":"%s"}', req_count, method); % sprintf() is much faster than strcat()
            else
                paramstr = '';
                for i = 1:length(varargin)
                    p = varargin{i};
                    if isnumeric(p)
                        if numel(p) == 1
                            paramstr = sprintf('%s%d,', paramstr, p);
                        elseif isnumeric(p) && numel(p) > 1
                            csv_str = sprintf('%d,', p(:));
                            paramstr = sprintf('%s[%s],', paramstr, csv_str(1:end-1));
                        end
                    elseif ischar(p)
                        paramstr = sprintf('%s"%s",', paramstr, p);
                    else
                        error('Unsupported parameter type.')
                    end
                end
                json_request = sprintf('{"jsonrpc":"2.0","id":%d,"method":"%s","params":[%s]}', ...
                    req_count, method, paramstr(1:end-1)...
                );
            end
            
            json_response = obj.httpRequest(json_request);
            
            if contains(json_response, 'error', 'IgnoreCase',true)
                resp = jsonToStruct(json_response);
                e = resp.error;
                error('\tERROR code: %d, MESSAGE: %s, DATA: %s', e.code, e.message, e.data);
            end
        end  
        
        
        function ret_str = httpRequest(obj, req)
            % Send HTTP request and return a response
            
            url = sprintf('http://%s:8050/json-rpc', obj.ip);
            ret_str = PulseStreamer.internal.urlread2(url, 'POST', req, []);
        end
        
        
        function resp = tcpRequest(obj, req)
            % Send request over TCP socket and return a response
            
            t = tcpclient(obj.ip, 21328);
            write(t, req.serialize());
            resp_bytes = read(t, 32);
            resp = PulseStreamer.internal.Response(resp_bytes);
        end
        
        
        function startCallbackTimer(obj)
            % Start timer that polls hasFinished state and executes
            % callback function if it was defined.
            
            % Before starting, make sure previous timer doesn't exist
            obj.stopCallbackTimer();
            
            % Create time with a name string that identifies timer ownership
            poll_timer = timer('Name', obj.cb_timer_name);
            poll_timer.TimerFcn = @obj.callbackTimerPollFcn;
            poll_timer.ErrorFcn = @(mt,evt) delete(mt); % if error occurs in timer it will delete itself
            poll_timer.Period = 0.1; %s
            poll_timer.ExecutionMode = 'fixedSpacing';
            start(poll_timer);
        end
        
        
        function callbackTimerPollFcn(obj, ~, ~)
            % Internal callback to check the status of the Pulse Streamer
            % and handling the external callback function.
            % The Pulse Streamer is polled only if the external callback function
            % was set by the user.

            try
                if obj.hasFinished()
                    obj.stopCallbackTimer();
                    obj.hasFinishedCallbackFcn(obj);
                end
            catch ME
                % cleanup the timer on error
                obj.stopCallbackTimer();
                rethrow(ME);
            end
        end
        
        
        function stopCallbackTimer(obj)
            %Stops and deletes timer object
            
            %Find timer by the name 
            tobj = timerfind('Name', obj.cb_timer_name);
            
            if isa(tobj, 'timer')
                if tobj.isvalid
                    stop(tobj);
                end
                delete(tobj)
            end
        end
        
        
        function checkFirmwareCompatibility(obj)
            % Checks if device firmware version is compatible with this
            % client implementation.
            
            % This checks for deprecated firmware version 0.9.
            json_resp = obj.httpRequest('{"jsonrpc":"2.0","id":"1","method":"getFirmwareVersion"}');
            if contains(json_resp, 'error', 'IgnoreCase', true) && ...
                    contains(json_resp, 'method not found', 'IgnoreCase', true)
                error(['This software requires Pulse Streamer firmware upgrade.\n' ...
                    'Please contact us:  %s'], obj.SUPPORT_EMAIL);
            end
            
            % Check if firmware version is within the supported range
            fwver_min = '1.0'; % lowest supported version
            fwver_max = '1.2'; % highest supported version
            fw_version = obj.getFirmwareVersion();
            obj.fw_ver = fw_version;
            
            switch PulseStreamer.internal.versionCheck(fwver_min, fw_version, fwver_max)
                case 1
                    error(['Newer, unsupported firmware version detected: "%s"\n' ...
                        'This firmware version requires client software update.\n' ...
                        'Please contact us:  %s'], ...
                        fw_version, support_email)
                case -1
                    error(['Unsupported firmware version detected: "%s"\n' ...
                        'This client software requires firmware version in range: ' ...
                        '"%s" <= version <= "%s"\n Please contact us:  %s'], ...
                        fw_version, fwver_min, fwver_max, obj.SUPPORT_EMAIL)
            end
            
            % This firmware version is supported
            % Capability checks
            
            % CHECK: version capabilities
            if strncmp(obj.fw_ver, '1.0', 3)
                obj.has_streamBinary = false;
                obj.can_streamBinary = false;
                warning('Old firmware detected. Please contact us for update instructions: %s', obj.SUPPORT_EMAIL);
            end
            
            % CHECK: if stream binary actually work
            if obj.has_streamBinary
                try
                    stream_cmd = PulseStreamer.internal.cmdStreamHighLevel();
                    resp = obj.callBinaryRPC(stream_cmd);
                    obj.can_streamBinary = (resp.status == PulseStreamer.internal.StatusEnum.OK);
                catch e
                    warning('Binary streaming is not available. Fallback to JSON stream.');
                    obj.can_streamBinary = false;
                end
            end
        end
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%% JSON-RPC PROTOCOL FUNCTIONS %%%%%%%%%%%


function stru = jsonToStruct(string)
    % converts the returned string from an JSON-RPC call 
    % to a struct
    
    stru = PulseStreamer.internal.parse_json(string);
    stru = stru{1};
end

function bool = jsonToBool(string)
    % converts the returned string from an JSON-RPC call to
    % a boolean value
    stru = jsonToStruct(string);
    if stru.result == 0
        bool = false;
    elseif stru.result == 1
        bool = true;
    else
        error(['return value is not a boolean (0 or 1) but ' num2str(stru.result)]);
    end
end

function value = jsonToUInt32(string)
    % extracts the returned value from an JSON-RPC call and
    % converts it into a uint32
    stru = jsonToStruct(string);
    value = uint32(stru.result);
end

function enc = base64EncodeJava(durations, digital, a0, a1)
    % Encode sequence data to base64 string using Java function

    N = numel(durations);
    % Cast each value into individual bytes. Reshape the result so the columns
    % represent the pulse and rows represent individual bytes for these
    % values. Number format: big-endian
    dur_bytes = reshape(typecast(swapbytes(uint32(durations(:))), 'int8'), 4, N);
    dig_bytes = reshape(typecast(swapbytes(uint8(digital(:))), 'int8'), 1, N);
    a0_bytes =  reshape(typecast(swapbytes(int16(a0(:))), 'int8'), 2, N);
    a1_bytes =  reshape(typecast(swapbytes(int16(a1(:))), 'int8'), 2, N);

    % combine byte arrays along byte dimension to get [9,N] array
    bytes = [dur_bytes; dig_bytes; a0_bytes; a1_bytes]; 
    % reshape byte arrays such that the corect byte sequence (along
    % columns) so the bytes are 
    % [b11,b21,b31,...,b91, b12,b22,b32,...,b92, ..., b1N,...,b9N]
    bytes = reshape(bytes, 1, 9 * N); 
    % encode into base64 string
    enc = transpose(char(org.apache.commons.codec.binary.Base64.encodeBase64(bytes, 0)));
end

function enc = base64EncodeMatlab(durations, digital, a0, a1)
    % Encode sequence data to base64 string using MATLAB function
    % Used as fallback when JVM is not available.

    N = numel(durations);
    % Cast each value into individual bytes. Reshape the result so the columns
    % represent the pulse and rows represent individual bytes for these
    % values. Number format: big-endian
    dur_bytes = reshape(typecast(swapbytes(uint32(durations(:))), 'uint8'), 4, N);
    dig_bytes = reshape(typecast(swapbytes(uint8(digital(:))), 'uint8'), 1, N);
    a0_bytes =  reshape(typecast(swapbytes(int16(a0(:))), 'uint8'), 2, N);
    a1_bytes =  reshape(typecast(swapbytes(int16(a1(:))), 'uint8'), 2, N);
    
    % combine byte arrays along byte dimension to get [9,N] array
    bytes = [dur_bytes; dig_bytes; a0_bytes; a1_bytes];
    
    % reshape byte arrays such that the corect byte sequence (along
    % columns) so the bytes are 
    % [b11,b21,b31,...,b91, b12,b22,b32,...,b92, ..., b1N,...,b9N]
    bytes = reshape(bytes, 1, 9 * N);
    
    % encode into base64 string
    enc = matlab.net.base64encode(bytes);
end


%%%%%%%%%%%%%%%%%%%%%%%%%%% BINARY PROTOCOL FUNCTIONS %%%%%%%%%%%

function bytes = encodeBinary(durations, digital, a0, a1)
    % Encode sequence data to base64 string using MATLAB function
    % Used as fallback when JVM is not available.

    N = numel(durations);
    % Cast each value into individual bytes. Reshape the result so the columns
    % represent the pulse and rows represent individual bytes for these
    % values. Number format: little-endian
    dur_bytes = reshape(typecast(uint32(durations(:)), 'uint8'), 4, N);
    dig_bytes = reshape(typecast(uint8(digital(:)), 'uint8'), 1, N);
    a0_bytes =  reshape(typecast(int16(a0(:)), 'uint8'), 2, N);
    a1_bytes =  reshape(typecast(int16(a1(:)), 'uint8'), 2, N);
    
    % combine byte arrays along byte dimension to get [9,N] array
    bytes = [dur_bytes; dig_bytes; a0_bytes; a1_bytes];
    
    % reshape byte arrays such that the corect byte sequence (along
    % columns) so the bytes are 
    % [b11,b21,b31,...,b91, b12,b22,b32,...,b92, ..., b1N,...,b9N]
    bytes = reshape(bytes, 1, 9 * N);
end

