classdef Sequence < handle
    %SEQUENCE class contains all information about the signal sequence
    % and the channel assignments. 
    %
    % https://www.swabianinstruments.com/static/documentation/PulseStreamer/sections/api-doc.html#sequence
    %
    
    properties (Access=private)
        
        achans = [];    % Analog channel index
        dchans = [];    % Digital channel index
        
        dp_time = {};   % Durations for each digital channel
        dp_level = {};  % Levels for each digital channel
        
        ap_time = {};   % Durations for each analog channel
        ap_level = {};  % Levels for each analog channel
    end
    
    methods
        
        function obj = Sequence(seq)
            %SEQUENCE Construct an instance of this class
            %    
            
            if exist('seq', 'var') 
                assert(isa(seq, 'PulseStreamer.Sequence'), 'Input parameter must be a Sequence object.');
                % copy object data
                obj = seq.copy();
            end 
        end
        
        function setDigital(obj, channels, pattern)
            %SETDIGITAL Maps signal pattern to digital output
            %   
            %    setDigital(channels, pattern)
            
            % Validate pattern parameter
            if ~(iscell(pattern) && size(pattern, 2) == 2)
                error('Pattern must be a cell array of dimensions [N,2]');
            end
            
            durs = cell2mat(pattern(:,1));
            lvls = cell2mat(pattern(:,2));
            
            assert(~any(durs<0), 'Negative duration values are not allowed in the pattern.');
            assert(~any(isnan(lvls)), 'NaN values are not allowed in the pattern.');
            assert(~any(isinf(lvls)), 'Inf values are not allowed in the pattern.');
            
            
            for chan = channels
                idx = find(chan == obj.dchans, 1);
                if isempty(idx)
                    obj.dchans(end+1) = chan;
                end
                idx = find(chan == obj.dchans);
                obj.dp_time{idx} = durs;
                obj.dp_level{idx} = double(lvls);
            end
        end
        
        function setAnalog(obj, channels, pattern)
            %SETANALOG Maps signal pattern to analog output
            %   
            
            % Validate pattern parameter
            if ~(iscell(pattern) && size(pattern, 2) == 2)
                error('Pattern must be a cell array of dimensions [N,2]');
            end
            
            durs = cell2mat(pattern(:,1));
            lvls = cell2mat(pattern(:,2));
            
            assert(~any(durs<0), 'Negative duration values are not allowed in the pattern.');
            assert(~any(isnan(lvls)), 'NaN values are not allowed in the pattern.');
            assert(~any(isinf(lvls)), 'Inf values are not allowed in the pattern.');
            
            for chan = channels
                idx = find(chan == obj.achans, 1);
                if isempty(idx)
                    obj.achans(end+1) = chan;
                end
                idx = find(chan == obj.achans);
                obj.ap_time{idx} = durs;
                obj.ap_level{idx} = double(lvls);
            end
        end
        
        function invertDigital(obj, channel)
           % Inverts digital pulse levels for the given channel
           
           idx = find(channel == obj.dchans);
           if ~isempty(idx)
               obj.dp_level{idx} = ~obj.dp_level{idx};
           end
        end
        
        function invertAnalog(obj, channel)
            % Inverts analog pulse levels for the given channel
            
            idx = find(channel == obj.achans);
            if ~isempty(idx)
                obj.ap_level{idx} = -obj.ap_level{idx};
            end
        end
        
        function TF = isempty(obj)
            % Returns True if sequence is empty
            
            % Find if any of the channels has
            dig_empty = all(cellfun(@(v) isempty(v), obj.dp_time));
            ana_empty = all(cellfun(@(v) isempty(v), obj.ap_time));
            TF = dig_empty & ana_empty;
        end
        
        function t_max = getDuration(obj)
            % Get sequence duration

            dur_dig = cellfun(@(v) sum(v), obj.dp_time);
            dur_ana = cellfun(@(v) sum(v), obj.ap_time);
            t_max = max([dur_dig(:); dur_ana(:)]);
            
        end
        
        function state = getLastState(obj)
            % GETLASTSTATE returns last state of the sequence.
            
            if obj.isempty()
                state = PulseStreamer.OutputState.ZERO;
            else
                d_msk = false(size(obj.dchans));
                for idx = 1:numel(obj.dp_level)
                   lvls = obj.dp_level{idx};
                   lvls = lvls(~isnan(lvls)); % Remove NaN values
                   d_msk(idx) = ~isempty(lvls) && logical(lvls(end));
                end
                
                a_lvls = zeros(size(obj.achans));
                for idx = 1:numel(obj.ap_level)
                    lvls = obj.ap_level{idx};
                    lvls = lvls(~isnan(lvls)); % Remove NaN values
                    if isempty(lvls)
                        a_lvls(idx) = 0;
                    else
                        a_lvls(idx) = lvls(end);
                    end
                end
                
                a0 = a_lvls(obj.achans == 0);
                a1 = a_lvls(obj.achans == 1);
                
                if isempty(a0); a0=0; end
                if isempty(a1); a1=0; end
                
                state = OutputState(obj.dchans(d_msk), a0, a1);
            end
        end
        
        function rle = getData(obj)
        % Get Sequence data
            
            % 1. Resolve KEEP flags where possible
            % 2. Remove zero durations and repetitions
            o = obj.sequenceWithKeepFlagsResolved(); 
            
            dpt = o.dp_time;
            dpl = o.dp_level;
            apt = o.ap_time;
            apl = o.ap_level;
            
            % find unique time points
            ct_dig = cell2mat(cellfun(@(v) cumsum(v), dpt(:), 'UniformOutput', false));
            ct_ana = cell2mat(cellfun(@(v) cumsum(v), apt(:), 'UniformOutput', false));
            ct_sync = unique([ct_dig(:); ct_ana(:)]);
            
            % allocate arrays
            Na = numel(o.achans); % Number of analog channels
            Nd = numel(o.dchans); % Number of digital channels
            Nst = numel(ct_sync); % number of sequence steps
            digi_vals = nan(Nst, Nd); 
            aout_vals = nan(Nst, Na);
            
            % resample digital channel levels for common time
            for chan = 1:Nd
                if ~isempty(dpl{chan})
                    ct = cumsum(dpt{chan});                    
                    lvls = dpl{chan};
                    digi_vals(:,chan) = resample_constpw(ct, lvls, ct_sync);
                end
            end
            
            % resample analog channel levels for common time
            for chan = 1:Na
                if ~isempty(apl{chan})
                    ct = cumsum(apt{chan});
                    lvls = apl{chan};
                    aout_vals(:,chan) = resample_constpw(ct, lvls, ct_sync);
                end
            end
           
            % Sort channel data by number ynd store to RLEData object 
            rle = PulseStreamer.internal.RLEData();
            rle.durations = diff([0; ct_sync]);
            rle.dchans = o.dchans;
            rle.dlvls = digi_vals;
            rle.achans = o.achans;
            rle.alvls = aout_vals; 
            
            % replace any remaining KEEP flags with default values
            % By now sequence is expected to have KEEP flag only where
            % default values are expected.
            rle.dlvls(isnan(rle.dlvls)) = 0; % Default value is zero
            rle.alvls(isnan(rle.alvls)) = 0; % Default value is zero
        end
        
        function out = repeat(obj, count)
            % REPEAT return sequence data duplicated "count" times.
            
            if count >= 0 && isscalar(count)
                
                out = obj.padded();
                
                % last repetition remains unpadded
                for idx = 1:numel(obj.dp_level)
                    out.dp_time{idx} = [repmat(out.dp_time{idx}, count-1, 1); obj.dp_time{idx}];
                    out.dp_level{idx} = [repmat(out.dp_level{idx}, count-1, 1); obj.dp_level{idx}];
                end
                
                for idx = 1:numel(obj.ap_level)
                    out.ap_time{idx} = [repmat(out.ap_time{idx}, count-1, 1); obj.ap_time{idx}];
                    out.ap_level{idx} = [repmat(out.ap_level{idx}, count-1, 1); obj.ap_level{idx}];
                end
            else
                error('Repeat "count" must be non-negative integer number');
            end
        end

        function plot(obj)
            % Plot both, analog and digital data in a current figure
            
            if obj.isempty()
                warning('Empty sequence can not be plotted.')
                return
            end
            clf();
            ax1=subplot(2,1,1);
            obj.plotAnalog();
            title('Analog outputs');
            legend show;
            ax2=subplot(2,1,2);
            obj.plotDigital();
            title('Digital outputs');
            xlabel('Time [ns]');
            linkaxes([ax1,ax2], 'x');
        end
        
        function plotDigital(obj)
            % Plot digital sequence data in current axes
            
            if obj.isempty()
                % Do not plot empty sequence
                return
            end
            
            o = obj.padded().sequenceWithKeepFlagsResolved();
            
            chlbl = arrayfun(@(v) {sprintf('D% 2d', v)}, o.dchans);
            ytick = 0:numel(o.dchans)-1;
            ax = gca;
            cla(ax);
            nextplot = ax.NextPlot; % Save original state
            ax.NextPlot = 'add';
            
            for ii = 1:numel(o.dchans)
                ct = [0; cumsum(o.dp_time{ii}(:))];
                lvls = o.dp_level{ii};
                lvls = [lvls(:); lvls(end)];
            
                % Plot zero line
                plot([ct(1), ct(end)], [1,1].*ytick(ii), ...
                    ':', 'Color', [0.3, 0.3, 0.3]);
                
                % Plot channel pattern
                ax.ColorOrderIndex = ii;
                stairs(ct, lvls*0.5+ytick(ii), ...
                    '.-', 'DisplayName', chlbl{ii}, 'LineWidth', 0.8,...
                    'MarkerSize', 8);
            end
            ax.NextPlot = nextplot; % Restore original state
            ax.YTick = ytick;
            ax.YTickLabel = chlbl;
            ylim(ax, [min(ytick)-0.2, max(ytick)+0.7]);
        end
        
        function plotAnalog(obj)
            % Plot analog sequence data in current axes
            
            if isempty(obj.achans)
                % plot nothing if no channels
                ylim([-1.05, 1.05]);
                return
            end
                        
            o = obj.padded().sequenceWithKeepFlagsResolved();
            
            chlbl = arrayfun(@(v) {sprintf('A% 2d', v)}, o.achans);
            ax = gca;
            cla(ax);
            nextplot = ax.NextPlot; % Save original state
            ax.NextPlot = 'add';
            for ii = 1:numel(o.achans)
                ct = [0; cumsum(o.ap_time{ii}(:))];
                lvls = o.ap_level{ii};
                lvls = [lvls(:); lvls(end)];
                lvls = resolve_keep_flag(lvls);
                
                % Plot channel pattern
                stairs(ct, lvls, ...
                    '.-', 'DisplayName', chlbl{ii}, 'LineWidth', 0.8,...
                    'MarkerSize', 8);
            end
            ax.NextPlot = nextplot; % Restore original state
            ylim([-1.05, 1.05]);
        end
            
        function newobj = copy(obj)
            % copy object data
            assert(isa(obj, 'PulseStreamer.Sequence'), 'Must be a Sequence object');
               
            class_constructor = str2func(class(obj)); % Get constructor 
            newobj = class_constructor(); % Instantiate 
            newobj.dchans = obj.dchans;
            newobj.dp_time = obj.dp_time;
            newobj.dp_level = obj.dp_level;
            
            newobj.achans = obj.achans;
            newobj.ap_time = obj.ap_time;
            newobj.ap_level = obj.ap_level;
        end
        
        function obj = horzcat(varargin)
            % Overrides horizontal concatenation
            obj = append(varargin{:});
        end
        
        function obj = vertcat(varargin)
            % Overrides vertical concatenation
            obj = append(varargin{:});
        end    
    end
    
    methods (Access={?PulseStreamer.Sequence})
        
        function objL = append(varargin)
            % APPEND Adds 'seq2' at the end of the 'seq1' sequence
            %   seq2    - Sequence object to be appended.
            
            isSeqObject = @(obj) isa(obj, 'PulseStreamer.Sequence');
            
            assert(all(cellfun(isSeqObject, varargin)), ...
                'All input objects must be of type "Sequence"');
            
            % Here we create object of exactly the same type as the first
            % This also properly handles subclasses of "Sequence"
            class_constructor = str2func('PulseStreamer.Sequence');
            objL = class_constructor(); % create left object
            
            % find inclusive set of channels 
            dchansL = unique(cell2mat(cellfun(@(o) o.dchans, varargin, 'UniformOutput', false)));
            achansL = unique(cell2mat(cellfun(@(o) o.achans, varargin, 'UniformOutput', false)));
            
            % initialize pattern storage
            objL.dchans = dchansL;
            objL.achans = achansL;
            objL.dp_time = cell(numel(dchansL), 1);
            objL.dp_level = cell(numel(dchansL), 1);
            objL.ap_time = cell(numel(achansL), 1);
            objL.ap_level = cell(numel(achansL), 1);
            
            for seq_i = 1:numel(varargin)
                objR = varargin{seq_i};
                
                % Concatenate digital channel data
                dchansR = objR.dchans;
                for iR = 1:numel(dchansR)
                    % find index of this channel in the left object
                    iL = find(dchansR(iR) == dchansL, 1); 
                    % Concatenate data
                    objL.dp_time{iL} = [objL.dp_time{iL}; objR.dp_time{iR}]; 
                    objL.dp_level{iL} = [objL.dp_level{iL}; objR.dp_level{iR}];
                end
                
                % Concatenate analog channel data
                achansR = objR.achans;
                for iR = 1:numel(achansR)
                    % find index of this channel in the left object
                    iL = find(achansR(iR) == achansL, 1); 
                    % Concatenate data
                    objL.ap_time{iL} = [objL.ap_time{iL}; objR.ap_time{iR}];
                    objL.ap_level{iL} = [objL.ap_level{iL}; objR.ap_level{iR}];
                end
                
                if seq_i < numel(varargin)
                    % Pad for every step except for the last 
                    objL = objL.padded();
                end
            end
        end
        
        function out = padded(obj)
            % Extends patterns in every channel to be of equal duration
            % Channels with no assignment are not padded and remain empty.
            % Empty channels will be filled with default value or latest
            % value when becomes available after concatenation.
            % 
            
            out = obj.copy();
            
            dp_dur = cellfun(@(v) sum(v), out.dp_time);
            ap_dur = cellfun(@(v) sum(v), out.ap_time);
            max_dur = max([dp_dur(:); ap_dur(:)]);
            dp_pad = max_dur - dp_dur;
            ap_pad = max_dur - ap_dur;
            
            for cidx = 1:numel(dp_pad)
                if dp_pad(cidx) < 1 % do not pad with 0 duration
                    continue
                end
                out.dp_time{cidx}(end+1,1) = dp_pad(cidx);
                out.dp_level{cidx}(end+1,1) = NaN;
            end
            
            for cidx = 1:numel(ap_pad)
                if ap_pad(cidx) < 1 % do not pad with 0 duration
                    continue
                end
                out.ap_time{cidx}(end+1,1) = ap_pad(cidx);
                out.ap_level{cidx}(end+1,1) = NaN;
            end 
        end
        
        function out = simplified(obj)
            % Returns object with simplified pattern data
            
            out = obj.copy();
            
            for ii = 1:numel(out.dchans)
                dur = out.dp_time{ii};
                lvl = out.dp_level{ii};
                [dur, lvl] = simplify_pattern(dur, lvl);
                out.dp_time{ii} = dur;
                out.dp_level{ii} = lvl;
            end
            for ii = 1:numel(out.achans)
                dur = out.ap_time{ii};
                lvl = out.ap_level{ii};
                [dur, lvl] = simplify_pattern(dur, lvl);
                out.ap_time{ii} = dur;
                out.ap_level{ii} = lvl;
            end
        end
        
        function out = sequenceWithKeepFlagsResolved(obj)
            % Resolves KEEP flags for all patterns
            out = obj.copy();
            
            for ii = 1:numel(out.dchans)
                lvl = out.dp_level{ii};
                lvl = resolve_keep_flag(lvl);
                out.dp_level{ii} = lvl;
            end
            for ii = 1:numel(out.achans)
                lvl = out.ap_level{ii};
                lvl = resolve_keep_flag(lvl);
                out.ap_level{ii} = lvl;
            end
        end
    end
end

function lvls = resolve_keep_flag(lvls)
    % Function resolves KEEP flags
    % Assume that KEEP flag has the NaN value

    % When first element is KEEP then the hardware default value will
    % be used. Class Sequence does not know what is the hardware
    % default so we preserve KEEP flag (NaN value). It shall be resolved
    % inside PulseStreamer class. 
    
    keep_msk = isnan(lvls);
    keep_msk(1) = false; % Exclude first even if "NaN"
    nan_idxs = find(keep_msk); % Indices of KEEP values
    for ii = 1:numel(nan_idxs)
        idx = nan_idxs(ii);
        lvls(idx) = lvls(idx-1); % Copy previous value
    end
end

function [durs, lvls] = simplify_pattern(durs, lvls)
% Simplifies pattern data by removing zero durations and combining
% repetitive levels. 

    % Find and remove zero durations from data except from last element.
    nz_msk = durs > 0;
    nz_msk(end) = true;
    durs = durs(nz_msk);
    lvls = lvls(nz_msk);

    % Find which values are repeated. First element is obviously excluded.
    val_rep_msk = [false; diff(lvls)==0]; % Repeated values
    nan_rep_msk = isnan(lvls) & isnan(circshift(lvls,1)); % Repeated NaN values
    rep_msk = val_rep_msk | nan_rep_msk;
    rep_msk(1) = false; % The first value can never be a repetition

    % Add up durations of repeated values.
    % For all repeated values, start from the end and add duration to previous
    % element. The first data element was already excluded before.
    rep_idxs = flipud(find(rep_msk(:))); % Start from the end!!!
    for ii = 1:numel(rep_idxs)
        idx = rep_idxs(ii);
        durs(idx-1) = durs(idx-1) + durs(idx);
    end

    % Get final result
    durs = durs(~rep_msk);
    lvls = lvls(~rep_msk);
end

function lvlq = resample_constpw(ct, lvl, ctq)
    % Resample using constant piecewise interpolation.
    % "ct" and "ctq" must be 
    % - ctq value such that ct(i) <= ctq < ct(i+1) -> lvlq = lvl(i);
    % - ctq value smaller than any value in ct -> lvlq = lvl(1);
    % - ctq value larger than any value in ct  -> lvlq = lvl(end);
    
    N = numel(ct);
    lvlq = zeros(numel(ctq), 1);
    jj=1;
    for ii=1:numel(ctq)
        while (ctq(ii)>ct(jj) && jj<N)
            jj=jj+1;
        end
        lvlq(ii)=lvl(jj);
    end
end
